// Library		libSOCK - Machine independent Socket I/O functions.
// File:		SOCK_Client.c
// Function:	SOCK_Client
// Descrizione: int SOCK_Client(char *hostn, int portn, int retry)
//				Apre un socket e lo connette alla porta/macchina date.
//				- hostn: Nome della macchina (/etc/hosts) a cui ci si
//				  vuole collegare.
//				- portn: numero di porta;
//				- retry: numero di ritentativi a distanza di RTT sec.
//				  1 sec per default, riassegnabile dall'esterno mediante
//				  SOCK_SetTout(). Se retry==0 => nessun ritentativo.
//				Connect() e` una chiamata sospensiva la cui durata, in
//				caso di mancata connessione dipende dai timeout interni
//				del TCP, tipicamente brevissimi. Talvolta pero` arrivano
//				a vari minuti (e non c'e`l'API per modificarli).
//				Quindi retry = n significa:
//				- n ritentativi istantanei intervallati da sleep di RTT
//				  sec, se il TCP ha timeout brevi, OPPURE
//				- un tentativo con timeout (alarm) di ((n+1) * RTT) sec,
//				  se il TCP e` configurato con dei timeout da WAN.
//				Dunque, comunque sia settato il TCP, il parametro "retry"
//				fissa la DURATA MASSIMA della chiamata.
//				Return: socket descr. (>0) se ha successo, -1 se fallisce.
// Author(s):	TECMAV srl - av/cb
// L.M. Date:	29-Nov-2003 (02:09:03)

#include "_libsock.h"
#include "libDPR.h"

static CML_SIG_HANDLER(ALRMTrap) { CML_Signal(SIGALRM, ALRMTrap); }


int SOCK_Client(char *hostn, int portn, int retry)
{
	struct sockaddr_in soc;		// Descr. ind. internet, per open socket.
	int sd1;					// Socket descriptor.
	int ret;					// Returned value.
	int siz;					// Size struttura "sockaddr_in"
	int RTT;					// Timeout base.
	int timeout;				// Max durata di connect se TCP ha timeout
								// lunghissimi.

	CMLsighp_t	PrevALRMTrap;	// Per restore SIGALRM handler.

	sd1   = 0;
	siz   = sizeof(struct sockaddr_in);
	if (retry < 0) retry = 0;
	RTT		= SOCK_GetTout();
	timeout = (retry+1) * RTT;


	DPR3("[ SOCK_Client(hostn=\"%s\", portn=%d, retry=%d)\n",
			hostn, portn, retry)

	do
	{
		if (sd1)
    	{
			if (sd1 > 0) close (sd1);

			if (errno == BAD_HOSTNAME || errno == EINTR)
			{
				DPR0("] -> -1\n")
				return(-1);
			}

			sleep (RTT);
			sd1 = 0;
		}

		errno = 0;
		memset (&soc, 0, siz);
		soc.sin_family	= AF_INET;
		SET_ADDR_LEN(soc)
		soc.sin_port = htons (portn);
		DPR_SOC(&soc)

		sd1 = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		DPR3("socket()=%d - errno=%d=%s\n", sd1, errno, CML_EStr())
		if (sd1 < 0) continue;


		if (SOCK_HAddrSI(hostn, 0, (int *)(&soc.sin_addr))) continue;
		DPR_SOC(&soc)

		PrevALRMTrap = CML_Signal(SIGALRM, ALRMTrap);
		alarm(timeout);

		ret = connect(sd1, (struct sockaddr *)&soc, sizeof(soc));
    	DPR4("connect(sd=%d)=%d - errno=%d=%s\n", sd1, ret, errno, CML_EStr())

		alarm(0);
		CML_Signal(SIGALRM, PrevALRMTrap);

    	if (!ret)
		{
			DPR1("] -> %d\n", sd1)
			return (sd1);
		}
	}
	while (retry--);

	if (sd1 > 0) close (sd1);
	DPR0("] -> -1\n")
	return (-1);
}
