// Modulo:		libSOC
// Funzione:	SOCK_Read()
// Autore:		A. Verardo
// Data U.M.:	12-Oct-2003 (19:48:20)
//
// Descrizione:	Legge con timeout il numero di caratteri richiesti e li
//				carica all'indirizzo indicato effettuando eventuali read
//				parziali.
//
// Sintassi:	int SOCK_Read(sd, pbuf, size, tout)
//	            	int  sd;
//	                char *pbuf;
//	                int  size, tout;
//
// Input:		sd:		socket id.
//				pbuf:	indirizzo del buffer utente. E` responsabilita`
//						del programma chiamante indicare un'area di almeno
//						"size" byte.
// 				size:	numero di caratteri da leggere.
//              tout:	durata max del completamento. Non vengono gestite
//						eventuali "alarm" gia` attive al momento della
//						chiamata. Se "tout" <= 0 non viene attivato alcun
//						meccanismo di timeout.
//
// Output:	 0:	quando sono stati effettivamente letti "syze" byte.
//			 1:	quando sono stati effettivamente letti "size" byte ed e`
//				stato riconosciuto l'arrivo di un signal non gestito da
//				questa procedura ma, evidentemente,	intrappolato a livello
//				piu` alto (vedi ret == -3).
//				Notare che quando "tout" <= 0, anche il SIGALRM e` conside-
//				rato "esterno" e provoca return 1 o -3.
//			-1:	in caso di errore di I/O (caduta del socket per terminazio-
//				ne del processo produttore o altro motivo, socket non aper-
//				to, parametri sbagliati etc.)
//			-2:	quando non e` stata effettuata alcuna lettura, a causa
//				dello scadere del timeout; possibile solo se "tout" > 0.
//			-3:	quando non e` stata effettuata alcuna lettura, a causa di
//				un signal esterno (vedi caso return == 1).
//
//  NOTE:		Se viene indicato "tout" > 0, qualunque sia l'esito finale,
//			    viene disabilitata la alarm, senza tener conto di eventuali
//				timeout preesistenti; la routine di gestione SIGALRM viene
//				invece ripristinata correttamente.
//
//				La variabile "errno" riflette sempre il reale andamento
//				della chiamata a meno dei casi:
//					(return -1) && (errno == 0) ===> socket down
//					(return -2) && (errno == 0) ===> timeout

#include "_libsock.h"
#include "libDPR.h"


#define SET_MY_TOUT		if(tout > 0)\
						{\
							prev = CML_Signal(SIGALRM, sig_alrm_handler);\
							alarm(tout);\
						}

#define RES_MY_TOUT		if(tout > 0){alarm(0); CML_Signal(SIGALRM, prev);}

static int sig_alrm_flag;

static CML_SIG_HANDLER(sig_alrm_handler)
{
	sig_alrm_flag = 1;
}


int SOCK_Read (int sd, char *pbuf, int size, int tout)
{
	CMLsighp_t prev;				// Ptr. a precedente SIGALRM handler.
	int letti;						// Byte letti ad ogni read.
	int retry;						// Contatore retry, usato come flag.

	DPR4("[ SOCK_Read(sd=%d, pbuf=0x%X, size=%d, tout=%d)\n",
		sd, (int)pbuf, size, tout)

	retry = errno = sig_alrm_flag = 0;			// Se fossero sporchi ...

	SET_MY_TOUT		 							// Set timeout.

	while ((letti = read(sd, pbuf, size)) != size)
	{
		if (sig_alrm_flag) 					// E` scattato il MIO timeout.
		{
			if (retry)						// Mi ha beccato a meta`
			{
				DPR3("  read %d bytes, errno=%d=%s -- continue\n",
					letti, errno, CML_EStr())
				letti = errno = sig_alrm_flag = 0;
			}
			else
			{ 								// Non aveva ancora letto nulla.
				RES_MY_TOUT
				DPR3("] read %d bytes, errno=%d=%s (reset) -- ret=-2\n",
					letti, errno, CML_EStr())
				errno = 0;
				return (-2);
			}
		}
		else
			if (letti <= 0)					// Problemi di connessione.
			{								// o signal "esterno".

				if (errno == EINTR)			// Signal esterno.
				{
					if (retry)				// Mi ha beccato a meta`
					{
						DPR3("  read %d bytes, errno=%d=%s -- continue\n",
							letti, errno, CML_EStr())
						letti = 0;
					}
					else
					{ 						// Non aveva ancora letto nulla.
						RES_MY_TOUT
						DPR3("] read %d bytes, errno=%d=%s -- ret=-3\n",
								letti, errno, CML_EStr())
						return (-3);
					}
				}
				else						// Socket down.
				{
					RES_MY_TOUT
					DPR3("] read %d bytes, errno=%d=%s -- ret=-1\n",
						letti, errno, CML_EStr())
					return (-1);
				}
			}

		DPR2(" retry %d/%d\n", letti, size)

		size -= letti;						// Lettura incompleta ...
		pbuf += letti;
		retry++;
	}

	DPR4("] read ok %d/%d errno=%d -- ret=%c\n",
		letti, size, errno, (errno ? '1' : '0'))

	RES_MY_TOUT

	return (errno ? 1 : 0);
}
