// Library		libSOCK - Machine independent Socket I/O functions.
// File:		SOCK_Server.c
// Functions:	SOCK_Server, SOCK_ServerF, SOCK_Listen
// Description:	See below.
// Author(s):	TECMAV srl - av/cb
// L.M. Date:	24-Sep-2003 (04:13:05)


#include "_libsock.h"
#include "libDPR.h"

// Funzione:	s_open_srv()
// Descrizione:	Implements SOCK_Server, SOCK_ServerF e SOCK_Listen
// 				depending on the value of the parameter "mode".

#define FORK	0
#define NOFORK	1
#define LISTEN	2

static HND_T AlrmTrap(int sig) {  }

static int s_open_srv (int portn, int retry, int mode, int timeout)
{
	struct sockaddr_in soc, soc_acc;	// Descr. indirizzo internet,
										// per apertura socket.
	int sd1, sd2;						// Socket descriptors.
	int siz;							// Size struct "sockaddr_in"
	int	ret;							// Returned code.
	int RTT;							// Timeout base.
	int tout;							// Copia di timeout per alarm.
	CMLsighp_t PrevAlrmTrap;			// Per restore SIGALRM handler
										// precedente.


	DPR4("[ s_open_srv(portn=%d, retry=%d, mode=%d, tout=%d)\n", 
			portn, retry, mode, timeout)
					
	RTT = SOCK_GetTout();

	if (timeout) PrevAlrmTrap = CML_Signal(SIGALRM, AlrmTrap);

	errno = 0; 
	sd1   = sd2 = 0;
	siz   = sizeof(struct sockaddr_in);
	if (retry < 0) retry = 0;

	tout = timeout;

	do
	{
		if (sd1)
		{
			if (sd1 > 0) close (sd1);
			if (sd2 > 0) close (sd2);

			if (ret < 0) 
			{
				sleep (RTT);
				if (timeout) tout = (tout > RTT ? tout-RTT : RTT);
			}
			sd1 = sd2 = 0;
		}
	
		errno = 0;
		memset (&soc, 0, siz);

		SET_ADDR_LEN(soc)
		soc.sin_family		= AF_INET;
		soc.sin_addr.s_addr	= INADDR_ANY;
		soc.sin_port		= htons(portn);
		DPR_SOC(&soc)


		sd1 = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		DPR6("socket(%d,%d,%d)=%d - errno=%d=%s\n",
			AF_INET, SOCK_STREAM, IPPROTO_TCP, sd1, errno, CML_EStr())
		if (sd1 < 0) continue;


		ret = 1;
		ret = setsockopt(sd1, SOL_SOCKET, SO_REUSEADDR, &ret, sizeof(ret));
		DPR2("setsockopt(sd=%d) (errno=%d)\n", sd1, errno)
		if (ret < 0) continue;

		ret = bind(sd1, (struct sockaddr *)(&soc), siz);
		DPR5("bind(sd=%d,..., siz=%d)=%d (errno=%d=%s)\n",
			sd1, siz, ret, errno, CML_EStr())
		DPR_SOC(&soc)
		if (ret) continue;


		ret = listen(sd1, 5);
		DPR3("listen(sd=%d)=%d (errno=%d)\n", sd1, ret, errno)
		if (ret < 0) continue;


		if (mode == LISTEN)
		{
			DPR1("] SOCK_Listen -> %d\n", sd1);
			return(sd1);
		}

		for (;;)
		{
			if (timeout) 
			{
				alarm(tout);
				DPR1("ALARM SET (%d sec)\n", tout)
			}

			DPR0("accept .....\n")
			memset (&soc_acc, 0, sizeof(soc_acc));
			sd2 = ret = accept(sd1,(struct sockaddr *)(&soc_acc),
											(SOCKL_T *)&siz);
			if (timeout) alarm(0);
			DPR5("accept(sd=%d,...,siz=%d)=%d (errno=%d) %s\n",
				sd1, siz, ret, errno, CML_EStr())
			DPR_SOC(&soc_acc)
			if (ret < 0) break;

			// Se deve forcare ed e` il padre ricomincia il for(;;) chiu-
			// dendo il socket a perdere, in tutti gli altri casi esce re-
			// stituendo l'ID del socket connesso.

#ifndef Os_ECOS

			// To be reviewed - ECOS doesn't have a POSIX fork()

			if (mode == FORK)
			{
				DPR2("fork. sd=%d al figlio, sd=%d chiuso\n", sd2, sd1)

				if (fork())
				{
					close(sd2);
					continue;
				}
			}
#endif

			close  (sd1);
			if (timeout) CML_Signal(SIGALRM, PrevAlrmTrap);
			DPR1("] -> %d\n", sd2);
			return (sd2);
		}
	}
	while (retry--);

	if (sd1 > 0) close (sd1);
	if (sd2 > 0) close (sd2);
	if (timeout) CML_Signal(SIGALRM, PrevAlrmTrap);
	DPR0("] close socket and return -1\n")
	return (-1);
}



//  Funzione:	SOCK_ServerF()
//  Autore:		A. Verardo
// L.M. Date:	27-Apr-2003 (00:02:53)
//
// Descrizione:	Apre un socket in modo "server" sulla porta indicata, e
//				mette in ascolto. Quando accetta una connessione, esegue
//				una "fork". Il processo padre continua ad ascoltare (non
//				esce mai dalla procedura); il processo figlio termina la
//				procedura restituendo il descrittore del socket connesso
//				al client "accettato" dal padre.
//
//Input:		portn: numero di porta;
//				retry: numero di ritentativi a distanza di 3 secondi,
//				in caso di errore ( retry <= 0 significa nessun ri-
//				tentativo).
//				Notare che esegue una accept() sospensiva, quindi
//				NON fa polling ogni RTT secondi: i retry vengono
//				spesi solo in caso di errore, con pause di RTT sec.
//
// Output:		socket file descriptor (>0) se ha successo, -1 se fallisce.

int SOCK_ServerF(int portn, int retry)
{
	return(s_open_srv(portn, retry, FORK, 0));
}


// Function:	SOCK_Server
// Description:	Apre un socket in modo "server" sulla porta indicata, e
//				mette in ascolto. Quando accetta una connessione,
//				restituisce il descrittore del socket connesso al client.
//  In/Out:		As above.

int SOCK_Server(int portn, int retry)
{
	return(s_open_srv(portn, retry, NOFORK, 0));
}



// Function:	SOCK_Listen
// Description:	Opens a socket for accept and returns it just after the
// 				listen() call. Useful for accpting connections on select()
// 				events.
//  In/Out:		As above.

int SOCK_Listen(int portn, int retry)
{
	return(s_open_srv(portn, retry, LISTEN, 0));
}
