// Funzione:	SOCK_Write()
// Autore:		A. Verardo
// Data U.M.:	19 Luglio 2001
//
// Descrizione:	Scrive con timeout il numero di caratteri richiesti prele-
//				vandoli dall'indirizzo indicato.
//				Se non riesce a soddisfare la richiesta con una singola
//				"write" di basso livello, effettua delle scritture parziali.
//
// Sintassi:	int SOCK_Write(sd, pbuf, size, tout)
//	                int  sd;
//	                char *pbuf;
//	                int  size, tout;
//
// Input:		sd:		socket id.
//				pbuf:	indirizzo del buffer utente. E` responsabilita`
//						del programma chiamante indicare un'area di almeno
//						"size" byte.
// 		 		size:	numero di caratteri da scrivere.
//              tout:	durata max del completamento. Non vengono gestite
//						eventuali "alarm" gia` attive al momento della
//						chiamata. Se "tout" <= 0 non viene attivato alcun
//						meccanismo interno di timeout.
//
// Output:	 0:	quando sono stati effettiv. scritti "syze" byte.
//			-1:	in caso di errore di I/O (caduta del socket per terminazio-
//				ne del processo produttore o altro motivo, socket non
//				aperto, parametri sbagliati etc.)
//			-2: in caso di timeout; possibile solo se "tout" > 0.
//			-3: se la scrittura e` stata interrotta da un signal non gesti-
//				to da questa procedura ma, evidentemente, intrappolato a
//				livello piu` alto.
//				Notare che quando "tout" <= 0, anche il SIGALRM	e` conside-
//				rato "esterno" e provoca return(-3).
//
//  Side Eff.:	Se viene indicato "tout" > 0, qualunque sia l'esito finale,
//			    viene disabilitata la alarm, senza tener conto di eventuali
//				timeout preesistenti; la routine di gestione SIGALRM viene
//				invece ripristinata correttamente.
//
//				La routine di gestione SIGPIPE viene sempre ripristinata.
//
//				La variabile "errno" riflette sempre il reale andamento
//				della chiamata a meno del caso:
//					(return -1) && (errno == EINTR) ===> socket down
//
//**************************************************************************

#include "_libsock.h"
#include "libDPR.h"


static int sig_pipe_flag;
static int sig_alrm_flag;

static CML_SIG_HANDLER(sig_pipe_handler) { sig_pipe_flag = 1; }

static CML_SIG_HANDLER(sig_alrm_handler) { sig_alrm_flag = 1; }


int SOCK_Write (int sd, char *pbuf, int size, int tout)
{
	CMLsighp_t	pspi;				// Ptr. a precedente SIGPIPE handler.
	CMLsighp_t	psal;				// Ptr. a precedente SIGALRM handler.
	int scritti;					// Byte scritti ad ogni write.

	errno = sig_alrm_flag = sig_pipe_flag = 0;	// Se fossero sporchi ...

	pspi  = CML_Signal(SIGPIPE, sig_pipe_handler);	// Save and set sig handler.

	if (tout > 0) 								// Event. set MIO timeout e
	{											// save precedente.
		psal = CML_Signal (SIGALRM, sig_alrm_handler);
		alarm(tout);
	}

	while ((scritti = write (sd, pbuf, size)) != size)
	{
		if (sig_alrm_flag) 					// E` scattato il MIO timeout.
		{
			CML_Signal (SIGALRM, psal);
			CML_Signal (SIGPIPE, pspi);
			DPR4("SOCK_Write: param. %d,%d,%d,%d\n",sd,(int)pbuf,size,tout)
			DPR3("SOCK_Write: write=%d (%d) %s, ret -2\n",
					scritti,errno,CML_EStr())
			return (-2);
		}

		if (scritti <  0)					// Problemi di connessione.
		{									// o write interrotta da qualche
			if (tout > 0) 					// signal gestito dall'esterno
			{								// (quando "tout" <= 0 anche il
				alarm(0);					// SIGALRM e` "esterno" !!!).
				CML_Signal (SIGALRM, psal);		// oppure parametri sbagliati.
			}

			CML_Signal (SIGPIPE, pspi);			// Comunque ripristina SIGPIPE.

			if ((errno == EINTR) && (!sig_pipe_flag))	// Signal esterno.
			{
				DPR4("SOCK_Write: param. %d,%d,%d,%d\n",sd,(int)pbuf,size,tout)
				DPR3("SOCK_Write: wr=%d (%d) %s,ret -3\n",scritti,errno,CML_EStr())
				return (-3);
			}
			
			DPR4("SOCK_Write: param. %d,%d,%d,%d\n",sd,(int)pbuf,size,tout)
			DPR3("SOCK_Write: wr=%d (%d) %s,ret -1\n",scritti,errno,CML_EStr())
			return (-1);
		}

		DPR2("SOCK_Write: retry %d/%d\n", scritti, size)

		size -= scritti;					// Scrittura incompleta ...
		pbuf += scritti;
	}

	if (tout > 0)							// Tutto ok.
	{
		alarm(0);
		CML_Signal (SIGALRM, psal);
	}

	CML_Signal (SIGPIPE, pspi);
	return (0);
}
